# -*- coding: utf-8 -*-
"""PREMIS-Reader-Writer: a small PREMIS library designed to work as a plugin
for METS-reader-writer. Public functions and classes:

- data_to_premis
- premis_to_data
- PREMISObject
- PREMISEvent
- PREMISAgent

"""
__metaclass__ = type

import abc
from datetime import datetime
import pprint
from uuid import uuid4

from lxml import etree
from lxml.builder import ElementMaker
import six

from . import utils


def now():
    return datetime.utcnow().replace(microsecond=0).isoformat("T")


def uuid():
    return str(uuid4())


@six.add_metaclass(abc.ABCMeta)
class PREMISElement:
    """Abstract base class for PREMIS object, event and agent classes. These
    classes must implement ``schema`` and ``defaults`` properties. After that,
    initalization can proceed either by passing a ``data`` kwarg to the class
    or by passing keyword arguments implicit in the element tag names of
    ``self.schema``, e.g.,::

        >>> premis_obj = PREMISObject(data=('object', {...}, (...)))
        >>> premis_obj = PREMISObject(
            identifier_type='UUID',
            identifier_value='9bf6bcf8-4d77-4623-a9fb-b703365d0ffe',
            ...)

    Under the first construction approach, the tuple passed as ``data`` becomes
    the source of truth for the PREMIS element. Under the second construction
    approach, the kwargs are used to construct a ``data`` tuple that becomes
    the source of truth. This tuple can be accessed via the ``.data`` property.
    """

    def __init__(self, **kwargs):
        self.premis_el_attrs = None
        self._attrs_to_paths = None
        self._attributes = None
        self._xml_element_values = {}
        self._xml_attribute_values = {}
        data = kwargs.get("data")
        if data:
            if isinstance(data, PREMISElement):
                data = data.data
            self._data = data
            self.premis_version = _premis_version_from_data(data)
        else:
            self.premis_version = kwargs.get("premis_version", utils.PREMIS_VERSION)
            self._xml_element_values = self._get_xml_element_values(kwargs)
            self._xml_attribute_values = _get_xml_attribute_values(
                kwargs, self.premis_version
            )
            self._data = self.generate_data()

    @property
    def data(self):
        return self._data

    @property
    def attrs_to_paths(self):
        """Return a dict that maps valid getter attributes to the simplified
        XPaths needed to get the corresponding values from ``self.data``.

        This property analyzes ``self.schema`` and sets
        ``self._attrs_to_paths`` to a dict that maps implicit getters like
        'agent_identifier_value' and 'identifier_value' to the XPaths implicit
        in ``self.schema``. In the case of ``PREMISAgent``, the above two
        getters would map to the XPath
        'agent/agent_identifier/agent_identifier_value'. ``PREMISAgent.schema``
        also implies the getters 'agent_identifier' and 'identifier', which
        both map to the XPath 'agent/agent_identifier' and which should return
        a tuple (or list thereof) instead of a string.
        """
        if self._attrs_to_paths:
            return self._attrs_to_paths
        self._attrs_to_paths = {}
        schema = self.schema
        tag = schema[0]
        attrs_to_paths_init = get_attrs_to_paths(schema)
        for attr, path in attrs_to_paths_init.items():
            if attr.startswith(tag + "_"):
                new_key = attr.replace(tag + "_", "", 1)
                self._attrs_to_paths[new_key] = path
            self._attrs_to_paths[attr] = path
            parts = path.split("/")[:-1]
            while parts:
                new_attr = parts[-1]
                new_path = "/".join(parts)
                self._attrs_to_paths[new_attr] = new_path
                parts = parts[:-1]
        return self._attrs_to_paths

    @property
    def attributes(self):
        """Return a dict that maps normalized XML attributes to their values,
        e.g., 'xsi_schema_location' and 'schema_location' would be keys for the
        value of the xsi:schemaLocation PREMIS XML attribute.
        """
        if self._attributes:
            return self._attributes
        self._attributes = {}
        for elem in self.data:
            if isinstance(elem, dict):
                for key, val in elem.items():
                    self._attributes[key] = val
                    if ":" in key:
                        key1 = key.replace(":", "_")
                        self._attributes[key1] = val
                        key2 = key.split(":", 1)[1]
                        self._attributes[key2] = val
        return self._attributes

    def serialize(self):
        return data_to_premis(self._data, self.premis_version)

    def tostring(self, pretty_print=True):
        return etree.tostring(self.serialize(), pretty_print=pretty_print)

    def __repr__(self):
        return repr(self._data)

    def __str__(self):
        return pprint.pformat(self._data, indent=4)

    def __eq__(self, other):
        """``self.data`` is the sole source of truth for ``PREMISElement``
        instances. Thus two such instances with the same data should compare
        equal as should an instance and a tuple if the instance's data is equal
        to the tuple.
        """
        if isinstance(other, PREMISElement):
            return self.data == other.data
        return self.data == other

    def __getattr__(self, attr_name):
        """Dynamically retrieve and return the value of an attribute which is
        implicitly defined by the return value of ``self.generate_data()``. All
        leaf node element names and full paths (with forward slashes replaced
        by double underscores) are now valid accessors. For example,
        ``premis_object.message_digest`` returns
        ``premis_object.findtext('object_characteristics/fixity/message_digest')``
        as does ``premis_object.object_characteristics__fixity__message_digest``.
        Similarly, ``premis_object.xsi_type``, ``premis_object.type`` and
        ``premis_object.xsi__type`` all return the value of the XML attribute
        xsi:type.
        """
        if attr_name in self.attrs_to_paths:
            return self.find_text_or_all(self.attrs_to_paths[attr_name])
        attr_name_norm = attr_name.replace("__", "/")
        if attr_name_norm in self.attrs_to_paths.values():
            return self.find_text_or_all(attr_name_norm)
        attr_name_norm = attr_name.replace("__", ":")
        if attr_name_norm in self.attributes:
            return self.attributes[attr_name_norm]
        valid_attributes = "\n".join(
            sorted(
                set(
                    list(self.attrs_to_paths.keys())
                    + [x.replace("/", "__") for x in self.attrs_to_paths.values()]
                    + list(x.replace(":", "_") for x in self.attributes.keys())
                )
            )
        )
        raise AttributeError(
            "Instance of {} has no attribute {}. Valid attributes"
            " are\n{}".format(self.__class__, attr_name, valid_attributes)
        )

    def find(self, path):
        return data_find(self._data, path)

    def findall(self, path):
        return data_find_all(self._data, path, dyn_cls=True)

    def findtext(self, path):
        return data_find_text(self._data, path)

    def find_text_or_all(self, path):
        return data_find_text_or_all(self._data, path, dyn_cls=True)

    @abc.abstractmethod
    def schema(self):
        """Return a tuple representing the schema of the PREMIS element.
        This tuple schema determines the available getters and setters (during
        initialization) of the subclass.
        """

    def generate_data(self):
        """Generate and return a tuple to assign to ``self._data``, which is
        the source of truth of the PREMIS XML element.
        Expects ``self._xml_element_values`` and ``self._xml_attribute_values`` to be dicts
        populated with XML element text values and XML attribute values,
        respectively.
        """
        return _generate_data(
            self.schema, self._xml_element_values, attributes=self._xml_attribute_values
        )

    @abc.abstractmethod
    def defaults(self):
        """Return a dict that maps implicit getter attributes (implicit in
        ``self.schema``) to default values or to callables that return default
        values. For example, see ``PREMISObject.defaults``.
        """

    def _get_xml_element_values(self, kwargs):
        """Using the user-supplied dict ``kwargs`` and the defaults returned by
        ``self.defaults``, return a dict mapping XML tag names (and paths) to
        values.
        """
        full_attrs_to_paths = {}
        xml_element_values = {}
        for attr_name, attr_path in self.attrs_to_paths.items():
            full_attrs_to_paths[attr_name] = attr_path
            full_attrs_to_paths[attr_path] = attr_path
        for attr_name, attr_path in full_attrs_to_paths.items():
            default = self.defaults.get(attr_name, self.defaults.get(attr_path, ""))
            if callable(default):
                default = default()
            val = kwargs.get(attr_name, kwargs.get(attr_path, default))
            xml_element_values[attr_name] = val
            xml_element_values[attr_path] = val
        return xml_element_values

    @classmethod
    def fromtree(cls, tree):
        """Create a PREMIS from an ``_Element``."""
        return cls(data=premis_to_data(tree))


class PREMISObject(PREMISElement):
    @property
    def defaults(self):
        return {
            "identifier_type": "UUID",
            "identifier_value": uuid,
            "composition_level": "1",
            "format_registry_name": "PRONOM",
            "date_created_by_application": now,
            "relationship": lambda: [],
            "inhibitors": lambda: [],
        }

    @property
    def schema(self):
        related_object_identifier, related_event_identifier = _get_relationship_tag_names(
            self.premis_version
        )
        return (
            "object",
            (
                "object_identifier",
                ("object_identifier_type",),
                ("object_identifier_value",),
            ),
            (
                "object_characteristics",
                ("composition_level",),
                ("fixity", ("message_digest_algorithm",), ("message_digest",)),
                ("size",),
                (
                    "format",
                    ("format_designation", ("format_name",), ("format_version",)),
                    (
                        "format_registry",
                        ("format_registry_name",),
                        ("format_registry_key",),
                    ),
                ),
                (
                    "creating_application",
                    ("creating_application_name",),
                    ("creating_application_version",),
                    ("date_created_by_application",),
                ),
                ("inhibitors", ("inhibitor_type",), ("inhibitor_target",)),
                ("object_characteristics_extension",),
            ),
            (
                "relationship",
                ("relationship_type",),
                ("relationship_sub_type",),
                (
                    related_object_identifier,
                    ("related_object_identifier_type",),
                    ("related_object_identifier_value",),
                ),
                (
                    related_event_identifier,
                    ("related_event_identifier_type",),
                    ("related_event_identifier_value",),
                ),
            ),
        )


class PREMISEvent(PREMISElement):
    @property
    def defaults(self):
        return {
            "identifier_type": "UUID",
            "identifier_value": uuid,
            "date_time": now,
            "linking_agent_identifier": lambda: [],
        }

    @property
    def schema(self):
        return (
            "event",
            (
                "event_identifier",
                ("event_identifier_type",),
                ("event_identifier_value",),
            ),
            ("event_type",),
            ("event_date_time",),
            ("event_detail",),
            (
                "event_outcome_information",
                ("event_outcome",),
                ("event_outcome_detail", ("event_outcome_detail_note",)),
            ),
            (
                "linking_agent_identifier",
                ("linking_agent_identifier_type",),
                ("linking_agent_identifier_value",),
            ),
        )

    @property
    def parsed_event_detail(self):
        """Parse and return our PREMIS eventDetail string value like::

            'program="7z"; version="9.20"; algorithm="bzip2"'

        and return a dict like::

            {'algorithm': 'bzip2', 'version': '9.20', 'program': '7z'}
        """
        return dict(
            [
                tuple([x.strip(' "') for x in kv.strip().split("=", 1)])
                for kv in self.event_detail.split(";")
            ]
        )

    # Compression Event Functionality
    # ==========================================================================

    @property
    def compression_details(self):
        """Return as a 3-tuple, this PREMIS compression event's program,
        version, and algorithm used to perform the compression.
        """
        event_type = self.findtext("event_type")
        if event_type != "compression":
            raise AttributeError(
                'PREMIS events of type "{}" have no compression'
                " details".format(event_type)
            )
        parsed_compression_event_detail = self.parsed_event_detail
        compression_program = _get_event_detail_attr(
            "program", parsed_compression_event_detail
        )
        compression_algorithm = _get_event_detail_attr(
            "algorithm", parsed_compression_event_detail
        )
        compression_program_version = _get_event_detail_attr(
            "version", parsed_compression_event_detail
        )
        archive_tool = {"7z": "7-Zip"}.get(compression_program, compression_program)
        return compression_algorithm, compression_program_version, archive_tool

    def get_decompression_transform_files(self, offset=0):
        """Returns a list of dicts representing ``<mets:transformFile>``
        elements with ``TRANSFORMTYPE="decompression"`` given
        ``compression_algorithm`` which is a comma-separated string of
        algorithms that must be used in the order provided to decompress
        the package, e.g., 'bzip2,tar' or 'lzma'.
        """
        compression_algorithm, _, _ = self.compression_details
        return [
            {
                "algorithm": algorithm,
                "order": str(index + offset + 1),
                "type": "decompression",
            }
            for index, algorithm in enumerate(compression_algorithm.split(","))
        ]

    # Encryption Event Functionality
    # ==========================================================================

    @property
    def encryption_details(self):
        """Return as a 3-tuple, this PREMIS encryption event's program,
        version, and key used to perform the encryption.
        """
        event_type = self.findtext("event_type")
        if event_type != "encryption":
            raise AttributeError(
                'PREMIS events of type "{}" have no encryption'
                " details".format(event_type)
            )
        parsed_encryption_event_detail = self.parsed_event_detail
        encryption_program = _get_event_detail_attr(
            "program", parsed_encryption_event_detail
        )
        encryption_program_version = _get_event_detail_attr(
            "version", parsed_encryption_event_detail
        )
        encryption_key = _get_event_detail_attr("key", parsed_encryption_event_detail)
        return encryption_program, encryption_program_version, encryption_key

    def get_decryption_transform_file(self):
        """Returns a dict representing a ``<mets:transformFile>`` element with
        ``TRANSFORMTYPE="decryption"``.
        """
        encryption_program, _, encryption_key = self.encryption_details
        return {
            "algorithm": encryption_program,
            "order": "1",
            "type": "decryption",
            "key": encryption_key,
        }


class PREMISAgent(PREMISElement):
    @property
    def defaults(self):
        return {}

    @property
    def schema(self):
        return (
            "agent",
            (
                "agent_identifier",
                ("agent_identifier_type",),
                ("agent_identifier_value",),
            ),
            ("agent_name",),
            ("agent_type",),
        )


def _data_to_lxml_el(data, ns, nsmap, element_maker=None, snake=True):
    """Convert tuple/list ``data`` to an ``lxml.etree._Element`` instance.
    :param tuple/list data: iterable whose first element is the snake-case
        string which is the name of the root XML element. Subsequent elements
        may be dicts (which encode XML attributes), tuples/lists (which encode
        sub-elements), or scalars (strings, ints or floats, which encode text
        under the element).
    :param str ns: the implicit namespace of all elements in the XML.
    :param dict nsmap: a dict of XML namespaces to define in the root element.
    :param ElementMaker element_maker: instance for creating XML elements.
    :returns: an ``lxml.etree._Element`` instance
    """
    if not element_maker:
        element_maker = ElementMaker(namespace=nsmap[ns], nsmap=nsmap)
    tag = data[0]
    if snake:
        camel_tag = utils.snake_to_camel(tag)
    func = getattr(element_maker, camel_tag)
    args = []
    attributes = {}
    for element in data[1:]:
        if isinstance(element, dict):
            for key, val in element.items():
                attributes[key] = val
        elif isinstance(element, (tuple, list)):
            args.append(
                _data_to_lxml_el(
                    element, ns, nsmap, element_maker=element_maker, snake=snake
                )
            )
        elif isinstance(element, six.text_type):
            args.append(element)
        elif isinstance(element, etree._Element):
            args.append(element)
        else:
            args.append(six.binary_type(element))
    ret = func(*args)
    for attr, val in attributes.items():
        try:
            ns, attr = attr.split(":")
        except ValueError:
            ns = None
        if snake:
            attr = utils.snake_to_camel(attr)
        if ns:
            attr = "{" + nsmap[ns] + "}" + attr
            ret.attrib[attr] = val
        else:
            ret.attrib[attr] = val
    return ret


def _to_colon_ns(bracket_ns, default_ns=None, nsmap=None, snake=True):
    """Convert a namespaced tag/attribute name from explicit XML "bracket"
    notation to a more succinct Pythonic colon-separated notation using
    snake_case, e.g.,::

        >>> _to_colon_ns(
            '{info:lc/xmlns/premis-v2}objectIdentifier',
            'premis', utils.NAMESPACES)
        'object_identifier'
        >>> _to_colon_ns('{info:lc/xmlns/premis-v2}objectIdentifier')
        'premis:object_identifier'
        >>> _to_colon_ns(
            'http://www.w3.org/2001/XMLSchema-instance}schemaLocation')
        'xsi:schema_location'
    """
    parts = [x.strip("{") for x in bracket_ns.split("}")]
    if len(parts) != 2:
        return bracket_ns
    ns, var = parts
    if default_ns and nsmap:
        try:
            ns = [k for k, v in nsmap.items() if v == ns][0]
            if ns == default_ns:
                if snake:
                    return utils.camel_to_snake(var)
                return var
        except IndexError:
            pass
    if snake:
        return ":".join([ns, utils.camel_to_snake(var)])
    return ":".join([ns, var])


def _get_el_attributes(lxml_el, ns=None, nsmap=None):
    """Return the XML attributes of lxml ``Element`` instance lxml_el as a dict
    where namespaced attributes are represented via colon-delimiting and using
    snake case.
    """
    attrs = {}
    for attr, val in lxml_el.items():
        attr = _to_colon_ns(attr, default_ns=ns, nsmap=nsmap)
        attrs[attr] = val
    return attrs


def _lxml_el_to_data(lxml_el, ns, nsmap, snake=True):
    """Convert an ``lxml._Element`` instance to a Python tuple."""
    tag_name = _to_colon_ns(lxml_el.tag, default_ns=ns, nsmap=nsmap)
    ret = [tag_name]
    attributes = _get_el_attributes(lxml_el, ns=ns, nsmap=nsmap)
    if attributes:
        ret.append(attributes)
    for sub_el in lxml_el:
        ret.append(_lxml_el_to_data(sub_el, ns, nsmap, snake=snake))
    text = lxml_el.text
    if text:
        ret.append(text)
    return tuple(ret)


def data_to_premis(data, premis_version=utils.PREMIS_VERSION):
    """Given tuple ``data`` representing a PREMIS entity (object, event or
    agent), return an ``lxml.etree._Element`` instance. E.g.,::

        >>> p = data_to_premis((
            'event',
            utils.PREMIS_META,
            (
                'event_identifier',
                ('event_identifier_type', 'UUID'),
                ('event_identifier_value', str(uuid4()))
            )
        ))
        >>> etree.tostring(p, pretty_print=True).decode('utf8')
        '''<premis:event
            xmlns:premis="info:lc/xmlns/premis-v2"
            xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
            version="2.2"
            xsi:schemaLocation="info:lc/xmlns/premis-v2 http://www.loc.gov/standards/premis/v2/premis-v2-2.xsd">
            <premis:eventIdentifier>
                <premis:eventIdentifierType>UUID</premis:eventIdentifierType>
                <premis:eventIdentifierValue>f4b7758f-e7b2-4155-9b56-d76965849fc1</premis:eventIdentifierValue>
            </premis:eventIdentifier>
        </premis:event>'''
    """
    nsmap = utils.PREMIS_VERSIONS_MAP[premis_version]["namespaces"]
    return _data_to_lxml_el(data, "premis", nsmap)


def premis_to_data(premis_lxml_el):
    """Transform a PREMIS ``lxml._Element`` instance to a Python tuple."""
    premis_version = premis_lxml_el.get("version", utils.PREMIS_VERSION)
    nsmap = utils.PREMIS_VERSIONS_MAP[premis_version]["namespaces"]
    return _lxml_el_to_data(premis_lxml_el, "premis", nsmap)


def data_find(data, path):
    """Find and return the first element-as-tuple in tuple ``data`` using simplified
    XPath ``path``.
    """
    path_parts = path.split("/")
    try:
        sub_elm = [
            el
            for el in data
            if isinstance(el, (tuple, list)) and el[0] == path_parts[0]
        ][0]
    except IndexError:
        return None
    else:
        if len(path_parts) > 1:
            return data_find(sub_elm, "/".join(path_parts[1:]))
        return sub_elm


def tuple_to_schema(tuple_):
    """Convert a tuple representing an XML data structure into a schema tuple
    that can be used in the ``.schema`` property of a sub-class of
    PREMISElement.
    """
    schema = []
    for element in tuple_:
        if isinstance(element, (tuple, list)):
            try:
                if isinstance(element[1], six.string_types):
                    schema.append((element[0],))
                else:
                    schema.append(tuple_to_schema(element))
            except IndexError:
                schema.append((element[0],))
        else:
            schema.append(element)
    return tuple(schema)


def generate_element_class(tuple_instance):
    """Dynamically create a sub-class of PREMISElement given
    ``tuple_instance``, which is a tuple representing an XML data structure.
    """
    schema = tuple_to_schema(tuple_instance)

    def defaults(self):
        return {}

    def schema_getter(self):
        return schema

    new_class_name = "PREMIS{}Element".format(schema[0].capitalize())
    return type(
        new_class_name,
        (PREMISElement,),
        {"defaults": property(defaults), "schema": property(schema_getter)},
    )


def data_find_all(data, path, dyn_cls=False):
    """Find and return all element-as-tuples in tuple ``data`` using simplified
    XPath ``path``.
    """
    path_parts = path.split("/")
    try:
        sub_elms = tuple(
            el
            for el in data
            if isinstance(el, (tuple, list)) and el[0] == path_parts[0]
        )
    except IndexError:
        return None
    if len(path_parts) > 1:
        ret = []
        for sub_elm in sub_elms:
            for x in data_find_all(sub_elm, "/".join(path_parts[1:])):
                ret.append(x)
        ret = tuple(ret)
    else:
        ret = sub_elms
    if ret and dyn_cls:
        cls = generate_element_class(ret[0])
        return tuple(cls(data=tuple_) for tuple_ in ret)
    return ret


def data_find_text(data, path):
    """Return the text value of the element-as-tuple in tuple ``data`` using
    simplified XPath ``path``.
    """
    el = data_find(data, path)
    if not isinstance(el, (list, tuple)):
        return None
    texts = [child for child in el[1:] if not isinstance(child, (tuple, list, dict))]
    if not texts:
        return None
    return " ".join(
        [
            # How should we deal with decoding errors when `x` is binary?
            # For now, we're using the ``strict`` mode. Other options here:
            # https://docs.python.org/3/library/functions.html#open.
            six.ensure_text(x, encoding="utf-8", errors="strict")
            for x in texts
        ]
    )


def data_find_text_or_all(data, path, dyn_cls=False):
    text = data_find_text(data, path)
    if text:
        return text
    return data_find_all(data, path, dyn_cls=dyn_cls)


def get_event_type(data):
    return data_find_text(data, "event_type")


def _get_event_detail_attr(attr, parsed_event_detail):
    try:
        return parsed_event_detail[attr]
    except KeyError:
        print(
            "Unable to find attribute {} in event detail {}".format(
                attr, parsed_event_detail
            )
        )
        return "No value found"


def _get_relationship_tag_names(premis_version):
    related_object_identifier = {"2.2": "related_object_identification"}.get(
        premis_version, "related_object_identifier"
    )
    related_event_identifier = {"2.2": "related_event_identification"}.get(
        premis_version, "related_event_identifier"
    )
    return related_object_identifier, related_event_identifier


def _generate_data(schema, elements, attributes=None, path=None):
    """Using tree-as-tuple ``schema`` as guide, return a tree-as-tuple ``data``
    representing a PREMIS XML element, where the values in dict ``elements`` and
    the values in dict ``attributes`` are located in the appropriate locations
    in the ``data`` tree structure.
    """
    path = path or []
    attributes = attributes or {}
    tag_name = schema[0]
    data = [tag_name]
    if attributes:
        data.append(attributes)
    new_path = path[:]
    new_path.append(tag_name)
    root = new_path[0]
    possible_paths = ["__".join(new_path), tag_name]
    if root != tag_name and tag_name.startswith(root):
        possible_paths.append(tag_name.lstrip(root)[1:])
    for possible_path in possible_paths:
        val = elements.get(possible_path)
        if val:
            if isinstance(val, (tuple, list)):
                data = tuple(val)
            else:
                if attributes:
                    data = (tag_name, attributes, val)
                else:
                    data = (tag_name, val)
            return tuple(data)
    for subschema in schema[1:]:
        subel = _generate_data(subschema, elements, path=new_path)
        if (not subel) or (subel == subschema):
            continue
        if all(map(lambda x: isinstance(x, tuple), subel)):
            for subsubel in subel:
                data.append(subsubel)
        elif not el_is_empty(subel):
            data.append(subel)
    return tuple(data)


def el_is_empty(el):
    """Return ``True`` if tuple ``el`` represents an empty XML element."""
    if len(el) == 1 and not isinstance(el[0], (list, tuple)):
        return True
    subels_are_empty = []
    for subel in el:
        if isinstance(subel, (list, tuple)):
            subels_are_empty.append(el_is_empty(subel))
        else:
            subels_are_empty.append(not bool(subel))
    return all(subels_are_empty)


def get_attrs_to_paths(schema, attrs_to_paths=None, path=None):
    """Analyze PREMIS-element-as-tuple ``schema`` and return a dict that maps
    attribute names to the simplified XPaths needed to retrieve them, e.g.,::

        >>> {'object_identifier_type':
                'object_identifier/object_identifier_type',
             'object_identifier_value':
                'object_identifier/object_identifier_value'}
    """
    attrs_to_paths = attrs_to_paths or {}
    tag = schema[0]
    if len(schema) == 1:
        attrs_to_paths[tag] = "/".join(path + [tag])
    else:
        for elem in schema[1:]:
            if isinstance(elem, dict):
                continue
            new_path = [] if path is None else path + [tag]
            if isinstance(elem, (list, tuple)):
                attrs_to_paths.update(
                    get_attrs_to_paths(
                        elem, attrs_to_paths=attrs_to_paths, path=new_path
                    )
                )
            else:
                attrs_to_paths[tag] = "/".join(new_path)
    return attrs_to_paths


def _get_xml_attribute_values(kwargs, premis_version=utils.PREMIS_VERSION):
    premis_el_attrs = utils.PREMIS_VERSIONS_MAP[premis_version]["meta"].copy()
    xsi_type = kwargs.get("xsi_type")
    if xsi_type:
        premis_el_attrs["xsi:type"] = xsi_type
    return premis_el_attrs


def _premis_version_from_data(data):
    """Given tuple ``data`` encoding a PREMIS element, attempt to return the
    PREMIS version it is using. If none can be found, return the default PREMIS
    version.
    """
    for child in data:
        if isinstance(child, dict):
            version = child.get("version")
            if version:
                return version
    return utils.PREMIS_VERSION



# -*- coding: utf-8 -*-
"""Dependency Injection logic for metsrw.

Here a global singleton feature broker is instantiated. By providing features
(i.e., dependencies) to the feature broker one can alter, for example, the
PREMIS dependencies used by metsrw's ``FSEntry`` class when reading and writing
PREMIS XML. Here we provide default classes for handling PREMIS using classes
defined in plugins.premisrw::

    >>> feature_broker = FeatureBroker()
    >>> feature_broker.provide('premis_object_class', premisrw.PREMISObject)
    >>> feature_broker.provide('premis_event_class', premisrw.PREMISEvent)
    >>> feature_broker.provide('premis_agent_class', premisrw.PREMISAgent)

However, the PREMIS dependencies can be altered by providing other features
with the same names, e.g.,::

    >>> feature_broker.provide('premis_object_class', MyOtherPREMISObjectClass)

See http://code.activestate.com/recipes/413268/
"""

from six import with_metaclass

from .plugins import premisrw


class FeatureBroker(object):
    """Feature broker allows for the provisioning of features. These features
    are dependencies that can be injected. Usage::

        >>> feature_broker = FeatureBroker()
        >>> feature_broker.provide('premis_object_class', premisrw.PREMISObject)
    """

    def __init__(self, allow_replace=True):
        self.providers = {}
        self.allow_replace = allow_replace

    def provide(self, feature_name, provider, *args, **kwargs):
        """Provide a feature named ``feature_name`` using the provider object
        ``provider`` and any arguments (``args``, ``kwargs``) needed by the
        provider if it is callable.
        """
        if not self.allow_replace:
            assert feature_name not in self.providers, "Duplicate feature: {!r}".format(
                feature_name
            )
        if callable(provider) and not isinstance(provider, type):
            self.providers[feature_name] = lambda: provider(*args, **kwargs)
        else:
            self.providers[feature_name] = lambda: provider

    def clear(self):
        self.providers.clear()

    def __len__(self):
        return len(self.providers)

    def __getitem__(self, feature_name):
        try:
            provider = self.providers[feature_name]
        except KeyError:
            raise KeyError("Unknown feature named {!r}".format(feature_name))
        return provider()


def set_feature_broker_to_default_state(fb):
    fb.clear()
    fb.provide("premis_object_class", premisrw.PREMISObject)
    fb.provide("premis_event_class", premisrw.PREMISEvent)
    fb.provide("premis_agent_class", premisrw.PREMISAgent)


feature_broker = FeatureBroker()  # global singleton feature broker
set_feature_broker_to_default_state(feature_broker)


class Dependency(object):
    """Non-overriding descriptor for declaring required dependencies in metsrw
    classes. In the following example usage the ``FSEntry`` class is declaring
    a dependency on a feature named 'premis_object_class' which is a class and
    which has methods ``fromtree`` and ``serialize``::

        >>> from .di import is_class, has_methods, Dependency
        >>> class FSEntry(object):
        ...     premis_object_class = Dependency(
        ...         has_methods('serialize'),
        ...         has_class_methods('fromtree'),
        ...         is_class)

    """

    def __init__(self, *assertions):
        self.dependency_name = None
        self.assertions = assertions

    def __get__(self, instance, owner):
        obj = feature_broker[self.dependency_name]
        for assertion in self.assertions:
            assert assertion(obj), (
                "The value {!r} of {!r} does not match the specified"
                " criteria".format(obj, self.dependency_name)
            )
        return obj


class DependencyPossessorMeta(type):
    """Metaclass for ``DependencyPossessor``. Classes that can have
    dependencies---i.e., classes that have class attributes whose values are
    instances of ``Dependency`` above---must inherit from
    ``DependencyPossessor``. This allows us to tell the ``Dependency`` instance
    that its ``dependency_name`` value should be the same as the name of the
    class attribute it was assigned to in the managed class. In short, it
    allows us to write ``premis_object_class = Dependency()`` instead of
    ``premis_object_class = Dependency('premis_object_class')``.
    """

    def __init__(cls, name, bases, attr_dict):
        super(DependencyPossessorMeta, cls).__init__(name, bases, attr_dict)
        for key, attr in attr_dict.items():
            if isinstance(attr, Dependency):
                attr.dependency_name = key


class DependencyPossessor(with_metaclass(DependencyPossessorMeta, object)):
    pass


# ==============================================================================
#   Assertions for declaring dependencies using Dependency
# ==============================================================================


def has_class_methods(*class_method_names):
    """Return a test function that, when given a class, returns ``True`` if that
    class has all of the class methods in ``class_method_names``. If an object
    is passed to the test function, check for the class methods on its
    class.
    """

    def test(cls):
        if not isinstance(cls, type):
            cls = type(cls)
        for class_method_name in class_method_names:
            try:
                class_method = getattr(cls, class_method_name)
                if class_method.__self__ is not cls:
                    return False
            except AttributeError:
                return False
        return True

    return test


def has_methods(*method_names):
    """Return a test function that, when given an object (class or an
    instance), returns ``True`` if that object has all of the (regular) methods
    in ``method_names``. Note: this is testing for regular methods only and the
    test function will correctly return ``False`` if an instance has one of the
    specified methods as a classmethod or a staticmethod. However, it will
    incorrectly return ``True`` (false positives) for classmethods and
    staticmethods on a *class*.
    """

    def test(obj):
        for method_name in method_names:
            try:
                method = getattr(obj, method_name)
            except AttributeError:
                return False
            else:
                if not callable(method):
                    return False
                if not isinstance(obj, type):
                    try:
                        # An instance method is a method type with a __self__
                        # attribute that references the instance.
                        if method.__self__ is not obj:
                            return False
                    except AttributeError:
                        return False
        return True

    return test


def is_class(obj):
    return isinstance(obj, type)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from itertools import chain
import logging
import os
from random import randint

from lxml import etree
import six

from .di import (
    is_class,
    has_methods,
    has_class_methods,
    Dependency,
    DependencyPossessor,
)
from . import exceptions
from .metadata import MDWrap, MDRef, SubSection, AMDSec
from . import utils

LOGGER = logging.getLogger(__name__)


class FSEntry(DependencyPossessor):
    """A class representing a filesystem entry - either a file or a directory.

    When passed to a :class:`metsrw.mets.METSDocument` instance, the tree of
    FSEntry objects will be used to construct the <fileSec> and <structMap>
    elements of a METS document.

    Unless otherwise specified, an FSEntry object is assumed to be a file;
    pass the `type` value as 'Directory' to specify that the object is
    instead a directory.

    An FSEntry object must be instantiated with a path as the first
    argument to the constructor, which represents its path on disk.

    An FSEntry object which is a Directory may have one or more children,
    representing files or directories contained within itself. Directory
    trees are designed for top-to-bottom traversal. Files cannot have
    children, and attempting to instantiate a file FSEntry object with
    children will raise a ValueError.

    Any FSEntry object may have one or more metadata entries associated
    with it; these can take the form of either references to other XML
    files on disk, which should be wrapped in MDRef objects, or
    wrapped copies of those XML files, which should be wrapped in
    MDWrap objects.

    :param str path: Path to the file on disk, as a bytestring. This will
        populate FLocat @xlink:href
    :param str label: Label in the structMap. If not provided, will be populated
        with the basename of path
    :param str use: Use for the fileGrp.  Items with identical uses will be
        grouped together.
    :param str type: Type of FSEntry this is. This will appear in the structMap.
    :param list children: List of :class:`metsrw.fsentry.FSEntry` that are
        direct children of this element in the structMap.  Only allowed if type
        is 'Directory'
    :param str file_uuid: UUID of this entry. Will be used to construct the
        FILEID used in the fileSec and structMap, and GROUPID.  Only required if
        type is 'Item'.
    :param metsrw.fsentry.FSEntry derived_from: FSEntry that this FSEntry is
        derived_from. This is used to set the GROUPID in the fileSec.
    :param str checksum: Value of the file's checksum. Required if checksumtype passed.
    :param str checksumtype: Type of the checksum. Must be one of
        :const:`FSEntry.ALLOWED_CHECKSUMS`.  Required if checksum passed.
    :param list transform_files: a list of dicts representing METS transform
        file elements, which provide "a means to access any subsidiary files
        listed below a <file> element by indicating the steps required to
        'unpack' or transform the subsidiary files."
    :raises ValueError: if children passed when type is not 'Directory'
    :raises ValueError: if only one of checksum or checksumtype passed
    :raises ValueError: if checksumtype is not in :const:`FSEntry.ALLOWED_CHECKSUMS`
    """

    ALLOWED_CHECKSUMS = (
        "Adler-32",
        "CRC32",
        "HAVAL",
        "MD5",
        "MNP",
        "SHA-1",
        "SHA-256",
        "SHA-384",
        "SHA-512",
        "TIGER WHIRLPOOL",
    )

    # Dependencies that must be injected. This means that an
    # ``FSEntry`` instance must be able to call ``self.premis_object_class`` and
    # get a class with methods ``fromtree`` and ``serialize``.
    premis_object_class = Dependency(
        has_methods("serialize"), has_class_methods("fromtree"), is_class
    )
    premis_event_class = Dependency(
        has_methods("serialize"), has_class_methods("fromtree"), is_class
    )
    premis_agent_class = Dependency(
        has_methods("serialize"), has_class_methods("fromtree"), is_class
    )

    PREMIS_OBJECT = "PREMIS:OBJECT"
    PREMIS_EVENT = "PREMIS:EVENT"
    PREMIS_AGENT = "PREMIS:AGENT"

    def __init__(
        self,
        path=None,
        label=None,
        use="original",
        type=u"Item",
        children=None,
        file_uuid=None,
        derived_from=None,
        checksum=None,
        checksumtype=None,
        transform_files=None,
        mets_div_type=None,
    ):
        # path can validly be any encoding; if this value needs
        # to be spliced later on, it's better to treat it as a
        # bytestring than as actually being encoded text.
        if six.PY2:
            if isinstance(path, six.text_type):
                self.path = path.encode("utf-8")
            else:
                self.path = path
        else:  # TODO: Py3 is still using Unicode.
            if isinstance(path, six.binary_type):
                self.path = path.decode("utf-8", errors="strict")
            else:
                self.path = path
        if label is None and path is not None:
            label = os.path.basename(path)
        self.label = label
        self.use = use
        self.type = six.text_type(type)
        self.parent = None
        self._children = []
        if not transform_files:
            transform_files = []
        self.transform_files = transform_files
        self.mets_div_type = mets_div_type or self.type
        children = children or []
        for child in children:
            self.add_child(child)
        self.file_uuid = file_uuid
        self.derived_from = derived_from
        if bool(checksum) != bool(checksumtype):
            raise ValueError(
                "Must provide both checksum and checksumtype, or neither. Provided values: %s and %s"
                % (checksum, checksumtype)
            )
        if checksumtype and checksumtype not in self.ALLOWED_CHECKSUMS:
            raise ValueError(
                "%s must be one of %s" % (checksumtype, self.ALLOWED_CHECKSUMS)
            )
        self.checksum = checksum
        self.checksumtype = checksumtype
        self.amdsecs = []
        self.dmdsecs = []

    @classmethod
    def dir(cls, label, children):
        """Return ``FSEntry`` directory object."""
        return FSEntry(label=label, children=children, type=u"Directory", use=None)

    @classmethod
    def from_fptr(cls, label, type_, fptr):
        """Return ``FSEntry`` object."""
        return FSEntry(
            label=label,
            type=type_,
            path=fptr.path,
            use=fptr.use,
            file_uuid=fptr.file_uuid,
            derived_from=fptr.derived_from,
            checksum=fptr.checksum,
            checksumtype=fptr.checksumtype,
        )

    def __str__(self):
        return "{s.type}: {s.path}".format(s=self)

    def __repr__(self):
        return "FSEntry(type={s.type!r}, path={s.path!r}, use={s.use!r}, label={s.label!r}, file_uuid={s.file_uuid!r}, checksum={s.checksum!r}, checksumtype={s.checksumtype!r})".format(
            s=self
        )

    # PROPERTIES

    def _create_id(self, prefix):
        return prefix + "_" + str(randint(1, 999999))

    def file_id(self):
        """ Returns the fptr @FILEID if this is not a Directory. """
        if self.type.lower() == "directory":
            return None
        if self.file_uuid is None:
            raise exceptions.MetsError(
                "No FILEID: File %s does not have file_uuid set" % self.path
            )
        if self.is_aip:
            return os.path.splitext(os.path.basename(self.path))[0]
        return utils.FILE_ID_PREFIX + self.file_uuid

    def group_id(self):
        """
        Returns the @GROUPID.

        If derived_from is set, returns that group_id.
        """
        if self.derived_from is not None:
            return self.derived_from.group_id()
        if self.file_uuid is None:
            return None
        return utils.GROUP_ID_PREFIX + self.file_uuid

    @property
    def admids(self):
        """ Returns a list of ADMIDs for this entry. """
        return [a.id_string() for a in self.amdsecs]

    @property
    def dmdids(self):
        """ Returns a list of DMDIDs for this entry. """
        return [d.id_string() for d in self.dmdsecs]

    @property
    def children(self):
        # Read-only
        return self._children

    @property
    def is_aip(self):
        return self.type.lower() == "archival information package"

    # ADD ATTRIBUTES

    def _add_metadata_element(self, md, subsection, mdtype, mode="mdwrap", **kwargs):
        """
        :param md: Value to pass to the MDWrap/MDRef
        :param str subsection: Metadata tag to create.  See :const:`SubSection.ALLOWED_SUBSECTIONS`
        :param str mdtype: Value for mdWrap/mdRef @MDTYPE
        :param str mode: 'mdwrap' or 'mdref'
        :param str loctype: Required if mode is 'mdref'. LOCTYPE of a mdRef
        :param str label: Optional. Label of a mdRef
        :param str otherloctype: Optional. OTHERLOCTYPE of a mdRef.
        :param str othermdtype: Optional. OTHERMDTYPE of a mdWrap.

        """
        # HELP how handle multiple amdSecs?
        # When adding *MD which amdSec to add to?
        if mode.lower() == "mdwrap":
            othermdtype = kwargs.get("othermdtype")
            mdsec = MDWrap(md, mdtype, othermdtype)
        elif mode.lower() == "mdref":
            loctype = kwargs.get("loctype")
            label = kwargs.get("label")
            otherloctype = kwargs.get("otherloctype")
            mdsec = MDRef(md, mdtype, loctype, label, otherloctype)
        subsection = SubSection(subsection, mdsec)
        if subsection.subsection == "dmdSec":
            self.dmdsecs.append(subsection)
        else:
            try:
                amdsec = self.amdsecs[0]
            except IndexError:
                amdsec = AMDSec()
                self.amdsecs.append(amdsec)
            amdsec.subsections.append(subsection)
        return subsection

    def add_techmd(self, md, mdtype, mode="mdwrap", **kwargs):
        return self._add_metadata_element(md, "techMD", mdtype, mode, **kwargs)

    def add_digiprovmd(self, md, mdtype, mode="mdwrap", **kwargs):
        return self._add_metadata_element(md, "digiprovMD", mdtype, mode, **kwargs)

    def add_rightsmd(self, md, mdtype, mode="mdwrap", **kwargs):
        return self._add_metadata_element(md, "rightsMD", mdtype, mode, **kwargs)

    def add_dmdsec(self, md, mdtype, mode="mdwrap", **kwargs):
        return self._add_metadata_element(md, "dmdSec", mdtype, mode, **kwargs)

    def serialize_md_inst(self, md_inst, md_class):
        """Serialize object ``md_inst`` by transforming it into an
        ``lxml.etree._ElementTree``. If it already is such, return it. If not,
        make sure it is the correct type and return the output of calling
        ``seriaize()`` on it.
        """
        valid_insts = tuple(
            chain((etree._ElementTree, etree._Element), six.string_types)
        )
        if isinstance(md_inst, valid_insts):
            return md_inst
        if not isinstance(md_inst, md_class):
            raise TypeError(
                "Instance {!r} must be instance of {!r}".format(md_inst, md_class)
            )
        return md_inst.serialize()

    def add_premis_object(self, md, mode="mdwrap"):
        meth = self.add_techmd
        if self.is_empty_dir:
            meth = self.add_dmdsec
        return meth(
            self.serialize_md_inst(md, self.premis_object_class),
            self.PREMIS_OBJECT,
            mode,
        )

    def add_premis_event(self, md, mode="mdwrap"):
        return self.add_digiprovmd(
            self.serialize_md_inst(md, self.premis_event_class), self.PREMIS_EVENT, mode
        )

    def add_premis_agent(self, md, mode="mdwrap"):
        return self.add_digiprovmd(
            self.serialize_md_inst(md, self.premis_agent_class), self.PREMIS_AGENT, mode
        )

    def add_premis_rights(self, md, mode="mdwrap"):
        # TODO add extra args and create PREMIS object here
        return self.add_rightsmd(md, "PREMIS:RIGHTS", mode)

    def add_dublin_core(self, md, mode="mdwrap"):
        # TODO add extra args and create DC object here
        return self.add_dmdsec(md, "DC", mode)

    def add_child(self, child):
        """Add a child FSEntry to this FSEntry.

        Only FSEntrys with a type of 'directory' can have children.

        This does not detect cyclic parent/child relationships, but that will
        cause problems.

        :param metsrw.fsentry.FSEntry child: FSEntry to add as a child
        :return: The newly added child
        :raises ValueError: If this FSEntry cannot have children.
        :raises ValueError: If the child and the parent are the same
        """
        if self.type.lower() != "directory":
            raise ValueError("Only directory objects can have children")
        if child is self:
            raise ValueError("Cannot be a child of itself!")
        if child not in self._children:
            self._children.append(child)
        child.parent = self
        return child

    def remove_child(self, child):
        """
        Remove a child from this FSEntry

        If `child` is not actually a child of this entry, nothing happens.

        :param child: Child to remove
        """
        try:
            self._children.remove(child)
        except ValueError:  # Child may not be in list
            pass
        else:
            child.parent = None

    # SERIALIZE

    def serialize_filesec(self):
        """
        Return the file Element for this file, appropriate for use in a fileSec.

        If this is not an Item or has no use, return None.

        :return: fileSec element for this FSEntry
        """
        if (
            self.type.lower() not in ("item", "archival information package")
            or self.use is None
        ):
            return None
        el = etree.Element(utils.lxmlns("mets") + "file", ID=self.file_id())
        if self.group_id():
            el.attrib["GROUPID"] = self.group_id()
        if self.admids:
            el.set("ADMID", " ".join(self.admids))
        if self.checksum and self.checksumtype:
            el.attrib["CHECKSUM"] = self.checksum
            el.attrib["CHECKSUMTYPE"] = self.checksumtype
        if self.path:
            flocat = etree.SubElement(el, utils.lxmlns("mets") + "FLocat")
            # Setting manually so order is correct
            try:
                flocat.set(utils.lxmlns("xlink") + "href", utils.urlencode(self.path))
            except ValueError:
                raise exceptions.SerializeError(
                    'Value "{}" (for attribute xlink:href) is not a valid'
                    " URL.".format(self.path)
                )
            flocat.set("LOCTYPE", "OTHER")
            flocat.set("OTHERLOCTYPE", "SYSTEM")
        for transform_file in self.transform_files:
            transform_file_el = etree.SubElement(
                el, utils.lxmlns("mets") + "transformFile"
            )
            for key, val in transform_file.items():
                attribute = "transform{}".format(key).upper()
                transform_file_el.attrib[attribute] = str(val)
        return el

    @property
    def is_empty_dir(self):
        """Returns ``True`` if this fs item is a directory with no children or
        a directory with only other empty directories as children.
        """
        if self.mets_div_type == "Directory":
            children = self._children
            if children:
                if all(child.is_empty_dir for child in children):
                    return True
                else:
                    return False
            else:
                return True
        else:
            return False

    def serialize_structmap(self, recurse=True, normative=False):
        """Return the div Element for this file, appropriate for use in a
        structMap.

        If this FSEntry represents a directory, its children will be
        recursively appended to itself. If this FSEntry represents a file, it
        will contain a <fptr> element.

        :param bool recurse: If true, serialize and apppend all children.
            Otherwise, only serialize this element but not any children.
        :param bool normative: If true, we are creating a "Normative Directory
            Structure" logical structmap, in which case we add div elements for
            empty directories and do not add fptr elements for files.
        :return: structMap element for this FSEntry
        """
        if not self.label:
            return None
        # Empty directories are not included in the physical structmap.
        if self.is_empty_dir and not normative:
            return None
        el = etree.Element(utils.lxmlns("mets") + "div", TYPE=self.mets_div_type)
        el.attrib["LABEL"] = self.label
        if (not normative) and self.file_id():
            etree.SubElement(el, utils.lxmlns("mets") + "fptr", FILEID=self.file_id())
        if self.dmdids:
            if (not normative) or (normative and self.is_empty_dir):
                el.set("DMDID", " ".join(self.dmdids))
        if recurse and self._children:
            for child in self._children:
                child_el = child.serialize_structmap(
                    recurse=recurse, normative=normative
                )
                if child_el is not None:
                    el.append(child_el)
        return el

    def get_subsections_of_type(self, mdtype, md_class):
        try:
            return [
                md_class.fromtree(ss.contents.document)
                for ss in self.amdsecs[0].subsections
                if ss.contents.mdtype == mdtype
            ]
        except IndexError:
            return []

    def get_premis_objects(self):
        return self.get_subsections_of_type(
            self.PREMIS_OBJECT, self.premis_object_class
        )

    def get_premis_events(self):
        return self.get_subsections_of_type(self.PREMIS_EVENT, self.premis_event_class)

    def get_premis_agents(self):
        return self.get_subsections_of_type(self.PREMIS_AGENT, self.premis_agent_class)

    def get_premis_event(self, event_uuid):
        try:
            return [
                evt
                for evt in self.get_premis_events()
                if evt.identifier_value == event_uuid
            ][0]
        except IndexError:
            return None



# -*- coding: utf-8 -*-
# Namespaces, etc.

XSI_NAMESPACE = "http://www.w3.org/2001/XMLSchema-instance"

# PREMIS v. 2.1
PREMIS_2_1_VERSION = "2.1"
PREMIS_2_1_NAMESPACE = "info:lc/xmlns/premis-v1"
PREMIS_2_1_XSD = "http://www.loc.gov/standards/premis/v2/premis-v2-1.xsd"
PREMIS_2_1_SCHEMA_LOCATION = "{} {}".format(PREMIS_2_1_NAMESPACE, PREMIS_2_1_XSD)
PREMIS_2_1_NAMESPACES = {"premis": PREMIS_2_1_NAMESPACE, "xsi": XSI_NAMESPACE}
PREMIS_2_1_META = {
    "xsi:schema_location": PREMIS_2_1_SCHEMA_LOCATION,
    "version": PREMIS_2_1_VERSION,
}

# PREMIS v. 2.2
PREMIS_2_2_VERSION = "2.2"
PREMIS_2_2_NAMESPACE = "info:lc/xmlns/premis-v2"
PREMIS_2_2_XSD = "http://www.loc.gov/standards/premis/v2/premis-v2-2.xsd"
PREMIS_2_2_SCHEMA_LOCATION = "{} {}".format(PREMIS_2_2_NAMESPACE, PREMIS_2_2_XSD)
PREMIS_2_2_NAMESPACES = {"premis": PREMIS_2_2_NAMESPACE, "xsi": XSI_NAMESPACE}
PREMIS_2_2_META = {
    "xsi:schema_location": PREMIS_2_2_SCHEMA_LOCATION,
    "version": PREMIS_2_2_VERSION,
}

# PREMIS v. 3.0
PREMIS_3_0_VERSION = "3.0"
PREMIS_3_0_NAMESPACE = "http://www.loc.gov/premis/v3"
PREMIS_3_0_XSD = "http://www.loc.gov/standards/premis/v3/premis.xsd"
PREMIS_3_0_SCHEMA_LOCATION = "{} {}".format(PREMIS_3_0_NAMESPACE, PREMIS_3_0_XSD)
PREMIS_3_0_NAMESPACES = {"premis": PREMIS_3_0_NAMESPACE, "xsi": XSI_NAMESPACE}
PREMIS_3_0_META = {
    "xsi:schema_location": PREMIS_3_0_SCHEMA_LOCATION,
    "version": PREMIS_3_0_VERSION,
}

PREMIS_VERSIONS_MAP = {
    PREMIS_2_1_VERSION: {"namespaces": PREMIS_2_2_NAMESPACES, "meta": PREMIS_2_1_META},
    PREMIS_2_2_VERSION: {"namespaces": PREMIS_2_2_NAMESPACES, "meta": PREMIS_2_2_META},
    PREMIS_3_0_VERSION: {"namespaces": PREMIS_3_0_NAMESPACES, "meta": PREMIS_3_0_META},
}

# Treat PREMIS v. 2.2 as the default (for now).
PREMIS_VERSION = PREMIS_2_2_VERSION
NAMESPACES = PREMIS_VERSIONS_MAP[PREMIS_VERSION]["namespaces"]
PREMIS_META = PREMIS_VERSIONS_MAP[PREMIS_VERSION]["meta"]
PREMIS_SCHEMA_LOCATION = PREMIS_2_2_SCHEMA_LOCATION


def lxmlns(arg, premis_version=PREMIS_VERSION):
    """Return XPath-usable namespace."""
    namespaces = PREMIS_VERSIONS_MAP[premis_version]["namespaces"]
    return "{" + namespaces[arg] + "}"


def snake_to_camel_cap(snake):
    """Convert snake_case to CamelCaseCapitalized."""
    return "".join([word.capitalize() for word in snake.split("_")])


def snake_to_camel(snake):
    """Convert snake_case to camelCase."""
    tmp = snake_to_camel_cap(snake)
    return tmp[0].lower() + tmp[1:]


def camel_to_snake(camel):
    """Convert camelCase to snake_case."""
    ret = []
    last_lower = False
    for char in camel:
        current_upper = char.upper() == char
        if current_upper and last_lower:
            ret.append("_")
            ret.append(char.lower())
        else:
            ret.append(char.lower())
        last_lower = not current_upper
    return "".join(ret)



# -*- coding: utf-8 -*-
import os

from lxml import etree, isoschematron
import six

from .utils import NAMESPACES

METS_XSD_PATH = "resources/mets.xsd"

# Right now there are two different schematron files for validating
# Archivematica-generated METS files vs Archivematica-generated METS pointer
# files. These could be consolidated to one.
AM_SCT_PATH = "resources/archivematica_mets_schematron.xml"
AM_PNTR_SCT_PATH = "resources/archivematica_mets_pointer_file_schematron.xml"


def _get_file_path(path):
    if not os.path.isfile(path):
        path_2 = os.path.join(os.path.dirname(os.path.abspath(__file__)), path)
        if not os.path.isfile(path_2):
            raise ValueError(
                "There is no (schema) file at either {} or {}".format(path, path_2)
            )
        return path_2
    return path


def get_schematron(sct_path):
    """Return an lxml ``isoschematron.Schematron()`` instance using the
    schematron file at ``sct_path``.
    """
    sct_path = _get_file_path(sct_path)
    parser = etree.XMLParser(remove_blank_text=True)
    sct_doc = etree.parse(sct_path, parser=parser)
    return isoschematron.Schematron(sct_doc, store_report=True)


def validate(mets_doc, xmlschema=METS_XSD_PATH, schematron=AM_SCT_PATH):
    """Validate a METS file using both an XMLSchema (.xsd) schema and a
    schematron schema, the latter of which typically places additional
    constraints on what a METS file can look like.
    """
    is_xsd_valid, xsd_error_log = xsd_validate(mets_doc, xmlschema=xmlschema)
    is_sct_valid, sct_report = schematron_validate(mets_doc, schematron=schematron)
    valid = is_xsd_valid and is_sct_valid
    report = {
        "is_xsd_valid": is_xsd_valid,
        "is_sct_valid": is_sct_valid,
        "xsd_error_log": xsd_error_log,
        "sct_report": sct_report,
    }
    report["report"] = report_string(report)
    return valid, report


def get_xmlschema(xmlschema, mets_doc):
    """Return a ``class::lxml.etree.XMLSchema`` instance given the path to the
    XMLSchema (.xsd) file in ``xmlschema`` and the
    ``class::lxml.etree._ElementTree`` instance ``mets_doc`` representing the
    METS file being parsed. The complication here is that the METS file to be
    validated via the .xsd file may reference additional schemata via
    ``xsi:schemaLocation`` attributes. We have to find all of these and import
    them from within the returned XMLSchema.

    For the solution that this is based on, see:
    http://code.activestate.com/recipes/578503-validate-xml-with-schemalocation/

    For other descriptions of the problem, see:
    - https://groups.google.com/forum/#!topic/archivematica/UBS1ay-g_tE
    - https://stackoverflow.com/questions/26712645/xml-type-definition-is-absent
    - https://stackoverflow.com/questions/2979824/in-document-schema-declarations-and-lxml
    """
    xsd_path = _get_file_path(xmlschema)
    xmlschema = etree.parse(xsd_path)
    schema_locations = set(
        mets_doc.xpath("//*/@xsi:schemaLocation", namespaces=NAMESPACES)
    )
    for schema_location in schema_locations:
        namespaces_locations = schema_location.strip().split()
        for namespace, location in zip(*[iter(namespaces_locations)] * 2):
            if namespace == NAMESPACES["mets"]:
                continue
            xs_import = etree.Element("{http://www.w3.org/2001/XMLSchema}import")
            xs_import.attrib["namespace"] = namespace
            xs_import.attrib["schemaLocation"] = location
            xmlschema.getroot().insert(0, xs_import)
    return etree.XMLSchema(xmlschema)


def xsd_validate(mets_doc, xmlschema=METS_XSD_PATH):
    xmlschema = get_xmlschema(xmlschema, mets_doc)
    is_valid = xmlschema.validate(mets_doc)
    error_log = xmlschema.error_log
    return is_valid, error_log


def schematron_validate(mets_doc, schematron=AM_SCT_PATH):
    """Validate a METS file using a schematron schema. Return a boolean
    indicating validity and a report as an ``lxml.ElementTree`` instance.
    """
    if isinstance(schematron, six.string_types):
        schematron = get_schematron(schematron)
    is_valid = schematron.validate(mets_doc)
    report = schematron.validation_report
    return is_valid, report


def sct_report_string(report):
    """Return a human-readable string representation of the error report
    returned by lxml's schematron validator.
    """
    ret = []
    namespaces = {"svrl": "http://purl.oclc.org/dsdl/svrl"}
    for index, failed_assert_el in enumerate(
        report.findall("svrl:failed-assert", namespaces=namespaces)
    ):
        ret.append(
            "{}. {}".format(
                index + 1,
                failed_assert_el.find("svrl:text", namespaces=namespaces).text,
            )
        )
        ret.append("   test: {}".format(failed_assert_el.attrib["test"]))
        ret.append("   location: {}".format(failed_assert_el.attrib["location"]))
        ret.append("\n")
    return "\n".join(ret)


def xsd_error_log_string(xsd_error_log):
    """Return a human-readable string representation of the error log
    returned by lxml's XMLSchema validator.
    """
    ret = []
    for error in xsd_error_log:
        ret.append(
            "ERROR ON LINE {}: {}".format(error.line, error.message.encode("utf-8"))
        )
    return "\n".join(ret)


def report_string(report):
    """Return a human-readable string representation of all of the validation
    errors.
    """
    return (
        "Schematron Error(s):\n"
        + sct_report_string(report["sct_report"])
        + "\n\nXMLSchema (xsd) Error(s):\n"
        + xsd_error_log_string(report["xsd_error_log"])
    )



# -*- coding: utf-8 -*-
"""
Classes for metadata sections of the METS. Include amdSec, dmdSec, techMD, rightsMD, sourceMD, digiprovMD, mdRef and mdWrap.
"""
from __future__ import absolute_import

import logging
from lxml import etree
from random import randint

import six

from . import exceptions
from . import utils

LOGGER = logging.getLogger(__name__)


class AMDSec(object):
    """
    An object representing a section of administrative metadata in a
    document.

    This is ordinarily created by :class:`metsrw.mets.METSDocument` instances and does not
    have to be instantiated directly.

    :param str section_id: ID of the section. If not provided, will be generated from 'amdSec' and a random number.
    :param list subsections: List of :class:`metsrw.metadata.SubSection` that are part of this amdSec
    :param Element tree: An lxml.Element that is an externally generated amdSec.  This will overwrite any automatic serialization.  If passed, section_id must also be passed.
    """

    tag = "amdSec"

    def __init__(self, section_id=None, subsections=None, tree=None):
        if subsections is None:
            subsections = []
        self.subsections = subsections
        self._id = section_id
        self._tree = tree
        if tree is not None and not section_id:
            raise ValueError("If tree is provided, section_id must also be provided")

    def id_string(self, force_generate=False):
        """
        Returns the ID string for the amdSec.

        :param bool force_generate: If True, will generate a new ID from 'amdSec' and a random number.
        """
        # e.g., amdSec_1
        if force_generate or not self._id:
            self._id = self.tag + "_" + str(randint(1, 999999))
        return self._id

    @classmethod
    def parse(cls, root):
        """
        Create a new AMDSec by parsing root.

        :param root: Element or ElementTree to be parsed into an object.
        """
        if root.tag != utils.lxmlns("mets") + "amdSec":
            raise exceptions.ParseError(
                "AMDSec can only parse amdSec elements with METS namespace."
            )
        section_id = root.get("ID")
        subsections = []
        for child in root:
            subsection = SubSection.parse(child)
            subsections.append(subsection)
        return cls(section_id, subsections)

    def serialize(self, now=None):
        """
        Serialize this amdSec and all children to lxml Element and return it.

        :param str now: Default value for CREATED in children if none set
        :return: amdSec Element with all children
        """
        if self._tree is not None:
            return self._tree
        el = etree.Element(utils.lxmlns("mets") + self.tag, ID=self.id_string())
        self.subsections.sort()
        for child in self.subsections:
            el.append(child.serialize(now))
        return el


class SubSection(object):
    """
    An object representing a metadata subsection in a document.

    This is usually created automatically and does not have to be instantiated directly.

    :param str subsection: Tag name for the subsection to be created.  Should be
        one of 'techMD', 'rightsMD', 'sourceMD' or 'digiprovMD' if contained in an
        :class:`amdSec`, or 'dmdSec'.
    :param contents: The MDWrap or MDRef contained in this subsection.
    :type contents: :class:`MDWrap` or :class:`MDRef`
    :param str section_id: ID of the section. If not provided, will be generated from subsection tag and a random number.
    """

    ALLOWED_SUBSECTIONS = ("techMD", "rightsMD", "sourceMD", "digiprovMD", "dmdSec")

    def __init__(self, subsection, contents, section_id=None):
        if subsection not in self.ALLOWED_SUBSECTIONS:
            raise ValueError(
                "%s must be one of %s" % (subsection, self.ALLOWED_SUBSECTIONS)
            )
        self.subsection = subsection
        self.contents = contents
        self._id = section_id
        self.status = None
        self.older = None
        self.newer = None
        self.created = None

    def __lt__(self, other):
        # Sort based on the subsection's order in ALLOWED_SUBSECTIONS
        # techMDs < rightsMD < sourceMD < digiprovMD < dmdSec
        return self.ALLOWED_SUBSECTIONS.index(
            self.subsection
        ) < self.ALLOWED_SUBSECTIONS.index(other.subsection)

    def id_string(self, force_generate=False):
        """
        Returns the ID string for this SubSection.

        :param bool force_generate: If True, will generate a new ID from the subsection tag and a random number.
        """
        if force_generate or not self._id:
            self._id = self.subsection + "_" + str(randint(1, 999999))
        return self._id

    def get_status(self):
        """
        Returns the STATUS when serializing.

        Calculates based on the subsection type and if it's replacing anything.

        :returns: None or the STATUS string.
        """
        if self.status is not None:
            return self.status
        if self.subsection == "dmdSec":
            if self.older is None:
                return "original"
            else:
                return "updated"
        if self.subsection in ("techMD", "rightsMD"):
            # TODO how to handle ones where newer has been deleted?
            if self.newer is None:
                return "current"
            else:
                return "superseded"
        return None

    def replace_with(self, new_subsection):
        """
        Replace this SubSection with new_subsection.

        Replacing SubSection must be the same time.  That is, you can only
        replace a dmdSec with another dmdSec, or a rightsMD with a rightsMD etc.

        :param new_subsection: Updated version of this SubSection
        :type new_subsection: :class:`SubSection`
        """
        if self.subsection != new_subsection.subsection:
            raise exceptions.MetsError(
                "Must replace a SubSection with one of the same type."
            )
        # TODO convert this to a DB so have bidirectonal foreign keys??
        self.newer = new_subsection
        new_subsection.older = self
        self.status = None

    @classmethod
    def parse(cls, root):
        """
        Create a new SubSection by parsing root.

        :param root: Element or ElementTree to be parsed into an object.
        :raises exceptions.ParseError: If root's tag is not in :const:`SubSection.ALLOWED_SUBSECTIONS`.
        :raises exceptions.ParseError: If the first child of root is not mdRef or mdWrap.
        """
        subsection = root.tag.replace(utils.lxmlns("mets"), "", 1)
        if subsection not in cls.ALLOWED_SUBSECTIONS:
            raise exceptions.ParseError(
                "SubSection can only parse elements with tag in %s with METS namespace"
                % (cls.ALLOWED_SUBSECTIONS,)
            )
        section_id = root.get("ID")
        created = root.get("CREATED", "")
        status = root.get("STATUS", "")
        child = root[0]
        if child.tag == utils.lxmlns("mets") + "mdWrap":
            mdwrap = MDWrap.parse(child)
            obj = cls(subsection, mdwrap, section_id)
        elif child.tag == utils.lxmlns("mets") + "mdRef":
            mdref = MDRef.parse(child)
            obj = cls(subsection, mdref, section_id)
        else:
            raise exceptions.ParseError(
                "Child of %s must be mdWrap or mdRef" % subsection
            )
        obj.created = created
        obj.status = status
        return obj

    def serialize(self, now=None):
        """
        Serialize this SubSection and all children to lxml Element and return it.

        :param str now: Default value for CREATED if none set
        :return: dmdSec/techMD/rightsMD/sourceMD/digiprovMD Element with all children
        """
        created = self.created if self.created is not None else now
        el = etree.Element(utils.lxmlns("mets") + self.subsection, ID=self.id_string())
        if created:  # Don't add CREATED if none was parsed
            el.set("CREATED", created)
        status = self.get_status()
        if status:
            el.set("STATUS", status)
        if self.contents:
            el.append(self.contents.serialize())
        return el


class MDRef(object):
    """
    An object representing an external XML document, typically associated
    with an :class:`metsrw.fsentry.FSEntry` object.

    :param str target: Path to the external document. MDRef does not validate
        the existence of this target.
    :param str mdtype: The string representing the mdtype of XML document being
        enclosed. Examples include "PREMIS:OBJECT" and "PREMIS:EVENT".
    :param str label: Optional LABEL for the mdRef element
    :param str loctype: LOCTYPE of the mdRef.  Must be one of 'ARK', 'URN', 'URL', 'PURL', 'HANDLE', 'DOI' or 'OTHER'.
    :param str otherloctype: OTHERLOCTYPE of the mdRef. Should be provided if loctype is OTHER.
    """

    VALID_LOCTYPE = ("ARK", "URN", "URL", "PURL", "HANDLE", "DOI", "OTHER")

    def __init__(self, target, mdtype, loctype, label=None, otherloctype=None):
        self.target = target
        self.mdtype = mdtype
        self.loctype = loctype
        if loctype not in self.VALID_LOCTYPE:
            raise ValueError(
                "loctype must be one of {}".format(", ".join(self.VALID_LOCTYPE))
            )
        self.label = label
        self.otherloctype = otherloctype

    @classmethod
    def parse(cls, root):
        """
        Create a new MDWrap by parsing root.

        :param root: Element or ElementTree to be parsed into a MDWrap.
        """
        if root.tag != utils.lxmlns("mets") + "mdRef":
            raise exceptions.ParseError(
                "MDRef can only parse mdRef elements with METS namespace."
            )
        # Required attributes
        mdtype = root.get("MDTYPE")
        if not mdtype:
            raise exceptions.ParseError("mdRef must have a MDTYPE")
        target = root.get(utils.lxmlns("xlink") + "href")
        if not target:
            raise exceptions.ParseError("mdRef must have an xlink:href.")
        try:
            target = utils.urldecode(target)
        except ValueError:
            raise exceptions.ParseError(
                'Value "{}" (of attribute xlink:href) is not a valid'
                " URL.".format(target)
            )
        loctype = root.get("LOCTYPE")
        if not loctype:
            raise exceptions.ParseError("mdRef must have a LOCTYPE")
        # Optional attributes
        label = root.get("LABEL")
        otherloctype = root.get("OTHERLOCTYPE")

        return cls(target, mdtype, loctype, label, otherloctype)

    def serialize(self):
        # If the source document is a METS document, the XPTR attribute of
        # this mdRef element should point to the IDs of each dmdSec element
        # in that document.
        XPTR = None
        try:
            target_doc = etree.parse(self.target)
            dmdsecs = [
                item.get("ID")
                for item in target_doc.findall(utils.lxmlns("mets") + "dmdSec")
            ]
            XPTR = "xpointer(id('{}'))".format(" ".join(dmdsecs))
        except Exception:
            pass

        el = etree.Element(utils.lxmlns("mets") + "mdRef")
        if self.label:
            el.attrib["LABEL"] = self.label
        if self.target:
            try:
                el.attrib[utils.lxmlns("xlink") + "href"] = utils.urlencode(self.target)
            except ValueError:
                raise exceptions.SerializeError(
                    'Value "{}" (for attribute xlink:href) is not a valid'
                    " URL.".format(self.target)
                )
        el.attrib["MDTYPE"] = self.mdtype
        el.attrib["LOCTYPE"] = self.loctype
        if self.otherloctype:
            el.attrib["OTHERLOCTYPE"] = self.otherloctype
        if XPTR:
            el.attrib["XPTR"] = XPTR
        return el


class MDWrap(object):
    """
    An object representing an XML document enclosed in a METS document.
    The entirety of the XML document will be included; to reference an
    external document, use the :class:`MDRef` class.

    :param str document: A string copy of the document, and will be parsed into
        an ElementTree at the time of instantiation.
    :param str mdtype: The MDTYPE of XML document being enclosed. Examples
        include "PREMIS:OBJECT", "PREMIS:EVENT,", "DC" and "OTHER".
    :param str othermdtype: The OTHERMDTYPE of the XML document. Should be set if mdtype is "OTHER".
    """

    def __init__(self, document, mdtype, othermdtype=None):
        parser = etree.XMLParser(remove_blank_text=True)
        if isinstance(document, six.string_types):
            self.document = etree.fromstring(document, parser=parser)
        elif isinstance(document, (etree._Element, list)):
            self.document = document
        self.mdtype = mdtype
        self.othermdtype = othermdtype

    @classmethod
    def parse(cls, root):
        """
        Create a new MDWrap by parsing root.

        :param root: Element or ElementTree to be parsed into a MDWrap.
        :raises exceptions.ParseError: If mdWrap does not contain MDTYPE
        :raises exceptions.ParseError: If xmlData contains no children
        """
        if root.tag != utils.lxmlns("mets") + "mdWrap":
            raise exceptions.ParseError(
                "MDWrap can only parse mdWrap elements with METS namespace."
            )
        mdtype = root.get("MDTYPE")
        if not mdtype:
            raise exceptions.ParseError("mdWrap must have a MDTYPE")
        othermdtype = root.get("OTHERMDTYPE")
        document = root.xpath("mets:xmlData/*", namespaces=utils.NAMESPACES)
        if len(document) == 0:
            raise exceptions.ParseError(
                "All mdWrap/xmlData elements must have at least one child; this"
                " one has none"
            )
        elif len(document) == 1:
            document = document[0]
        return cls(document, mdtype, othermdtype)

    def serialize(self):
        el = etree.Element(utils.lxmlns("mets") + "mdWrap", MDTYPE=self.mdtype)
        if self.othermdtype:
            el.attrib["OTHERMDTYPE"] = self.othermdtype
        xmldata = etree.SubElement(el, utils.lxmlns("mets") + "xmlData")
        if isinstance(self.document, list):
            for child in self.document:
                xmldata.append(child)
        else:
            xmldata.append(self.document)

        return el



# -*- coding: utf-8 -*-
"""PREMIS reader and writer."""

from __future__ import absolute_import
import logging

from .premis import (
    PREMISElement,
    PREMISObject,
    PREMISEvent,
    PREMISAgent,
    data_to_premis,
    premis_to_data,
    data_find,
    data_find_all,
    data_find_text,
    data_find_text_or_all,
)
from .utils import (
    XSI_NAMESPACE,
    PREMIS_2_1_VERSION,
    PREMIS_2_1_NAMESPACE,
    PREMIS_2_1_XSD,
    PREMIS_2_1_SCHEMA_LOCATION,
    PREMIS_2_1_NAMESPACES,
    PREMIS_2_1_META,
    PREMIS_2_2_VERSION,
    PREMIS_2_2_NAMESPACE,
    PREMIS_2_2_XSD,
    PREMIS_2_2_SCHEMA_LOCATION,
    PREMIS_2_2_NAMESPACES,
    PREMIS_2_2_META,
    PREMIS_3_0_VERSION,
    PREMIS_3_0_NAMESPACE,
    PREMIS_3_0_XSD,
    PREMIS_3_0_SCHEMA_LOCATION,
    PREMIS_3_0_NAMESPACES,
    PREMIS_3_0_META,
    PREMIS_VERSIONS_MAP,
    PREMIS_VERSION,
    NAMESPACES,
    PREMIS_META,
    PREMIS_SCHEMA_LOCATION,
    lxmlns,
    snake_to_camel_cap,
    snake_to_camel,
    camel_to_snake,
)


LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

__all__ = [
    "PREMISElement",
    "PREMISObject",
    "PREMISEvent",
    "PREMISAgent",
    "data_to_premis",
    "premis_to_data",
    "data_find",
    "data_find_all",
    "data_find_text",
    "data_find_text_or_all",
    "XSI_NAMESPACE",
    "PREMIS_2_1_VERSION",
    "PREMIS_2_1_NAMESPACE",
    "PREMIS_2_1_XSD",
    "PREMIS_2_1_SCHEMA_LOCATION",
    "PREMIS_2_1_NAMESPACES",
    "PREMIS_2_1_META",
    "PREMIS_2_2_VERSION",
    "PREMIS_2_2_NAMESPACE",
    "PREMIS_2_2_XSD",
    "PREMIS_2_2_SCHEMA_LOCATION",
    "PREMIS_2_2_NAMESPACES",
    "PREMIS_2_2_META",
    "PREMIS_3_0_VERSION",
    "PREMIS_3_0_NAMESPACE",
    "PREMIS_3_0_XSD",
    "PREMIS_3_0_SCHEMA_LOCATION",
    "PREMIS_3_0_NAMESPACES",
    "PREMIS_3_0_META",
    "PREMIS_VERSIONS_MAP",
    "PREMIS_VERSION",
    "NAMESPACES",
    "PREMIS_META",
    "PREMIS_SCHEMA_LOCATION",
    "lxmlns",
    "snake_to_camel_cap",
    "snake_to_camel",
    "camel_to_snake",
]



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from builtins import object
from collections import OrderedDict, namedtuple
from datetime import datetime
import logging
import os
import six
import sys

from lxml import etree

# This package
from . import exceptions
from . import fsentry
from . import metadata
from . import utils


LOGGER = logging.getLogger(__name__)

AIP_ENTRY_TYPE = "archival information package"
FPtr = namedtuple(
    "FPtr", "file_uuid derived_from use path amdids checksum checksumtype"
)


class METSDocument(object):
    def __init__(self):
        # Stores the ElementTree if this was parsed from an existing file
        self.tree = None
        # Only root-level elements are stored, since the rest
        # can be inferred via their #children attribute
        self.createdate = None
        self.objid = None
        self._root_elements = []
        self._all_files = None
        self._iter = None
        self.dmdsecs = []
        self.amdsecs = []

    @classmethod
    def read(cls, source):
        """Read ``source`` into a ``METSDocument`` instance. This is an
        instance constructor. The ``source`` may be a path to a METS file, a
        file-like object, or a string of XML.
        """
        if hasattr(source, "read"):
            return cls.fromfile(source)
        if os.path.exists(source):
            return cls.fromfile(source)
        if isinstance(source, six.string_types):
            source = source.encode("utf8")
        return cls.fromstring(source)

    # FSENTRYS

    def _collect_all_files(self, files=None):
        """
        Collect all FSEntrys into a set, including all descendants.

        :param list files: List of :class:`FSEntry` to traverse.
        :returns: Set of FSEntry
        """
        if files is None:
            files = self._root_elements
        collected = set()
        for entry in files:
            collected.add(entry)
            collected.update(self._collect_all_files(entry.children))
        return collected

    def all_files(self):
        """
        Return a set of all FSEntrys in this METS document.

        :returns: Set containing all :class:`FSEntry` in this METS document,
            including descendants of ones explicitly added.
        """
        # FIXME cache this. Should not break when add_child is called on an
        # element already in the document.
        return self._collect_all_files(self._root_elements)

    def get_file(self, **kwargs):
        """
        Return the FSEntry that matches parameters.

        :param str file_uuid: UUID of the target FSEntry.
        :param str label: structMap LABEL of the target FSEntry.
        :param str type: structMap TYPE of the target FSEntry.
        :returns: :class:`FSEntry` that matches parameters, or None.
        """
        # TODO put this in a sqlite DB so it can be queried efficiently
        # TODO handle multiple matches (with DB?)
        # TODO check that kwargs are actual attrs
        for entry in self.all_files():
            if all(value == getattr(entry, key) for key, value in kwargs.items()):
                return entry
        return None

    def append_file(self, fs_entry):
        """
        Adds an FSEntry object to this METS document's tree. Any of the
        represented object's children will also be added to the document.

        A given FSEntry object can only be included in a document once,
        and any attempt to add an object the second time will be ignored.

        :param metsrw.mets.FSEntry fs_entry: FSEntry to add to the METS document
        """

        if fs_entry in self._root_elements:
            return
        self._root_elements.append(fs_entry)
        # Reset file lists so they get regenerated with the new files(s)
        self._all_files = None

    append = append_file

    def remove_entry(self, fs_entry):
        """Removes an FSEntry object from this METS document.

        Any children of this FSEntry will also be removed. This will be removed
        as a child of it's parent, if any.

        :param metsrw.mets.FSEntry fs_entry: FSEntry to remove from the METS
        """
        try:
            self._root_elements.remove(fs_entry)
        except ValueError:  # fs_entry may not be in the root elements
            pass
        if fs_entry.parent:
            fs_entry.parent.remove_child(fs_entry)
        # Reset file lists so they get regenerated without the removed file(s)
        self._all_files = None

    remove = remove_entry

    # The following methods allow us to iterate over the FSEntry instances of a
    # METSDocument---``for fsentry in mets: ...``---, count
    # them---``len(mets)``---, and fetch them by
    # index---``my_fsentry = mets[21]``.

    def __iter__(self):
        return self

    def __len__(self):
        return len(self.all_files())

    def _get_all_files_list(self):
        return sorted(self.all_files(), key=lambda fsentry: fsentry.path or "")

    def __getitem__(self, index):
        return self._get_all_files_list()[index]

    def __next__(self):  # Py3-style iterator interface
        if self._iter is None:
            self._iter = iter(self._get_all_files_list())
        return next(self._iter)

    # SERIALIZE

    def _document_root(self, fully_qualified=True):
        """
        Return the mets Element for the document root.
        """
        nsmap = {"xsi": utils.NAMESPACES["xsi"], "xlink": utils.NAMESPACES["xlink"]}
        if fully_qualified:
            nsmap["mets"] = utils.NAMESPACES["mets"]
        else:
            nsmap[None] = utils.NAMESPACES["mets"]
        attrib = {
            "{}schemaLocation".format(utils.lxmlns("xsi")): utils.SCHEMA_LOCATIONS
        }
        if self.objid:
            attrib["OBJID"] = self.objid
        return etree.Element(utils.lxmlns("mets") + "mets", nsmap=nsmap, attrib=attrib)

    def _mets_header(self, now):
        """
        Return the metsHdr Element.
        """
        if self.createdate is None:
            e = etree.Element(utils.lxmlns("mets") + "metsHdr", CREATEDATE=now)
        else:
            e = etree.Element(
                utils.lxmlns("mets") + "metsHdr",
                CREATEDATE=self.createdate,
                LASTMODDATE=now,
            )
        return e

    @staticmethod
    def _collect_mdsec_elements(files):
        """
        Return all dmdSec and amdSec classes associated with the files.

        Returns all dmdSecs, then all amdSecs, so they only need to be
        serialized before being appended to the METS document.

        :param List files: List of :class:`FSEntry` to collect MDSecs for.
        :returns: List of AMDSecs and SubSections
        """
        dmdsecs = []
        amdsecs = []
        for f in files:
            for d in f.dmdsecs:
                dmdsecs.append(d)
            for a in f.amdsecs:
                amdsecs.append(a)

        dmdsecs.sort(key=lambda x: x.id_string())
        amdsecs.sort(key=lambda x: x.id_string())
        return dmdsecs + amdsecs

    def _structmap(self):
        """
        Returns structMap element for all files.
        """
        structmap = etree.Element(
            utils.lxmlns("mets") + "structMap",
            TYPE="physical",
            # TODO Add ability for multiple structMaps
            ID="structMap_1",
            # TODO don't hardcode this
            LABEL="Archivematica default",
        )
        for item in self._root_elements:
            child = item.serialize_structmap(recurse=True)
            if child is not None:
                structmap.append(child)

        return structmap

    def _normative_structmap(self):
        """Returns the normative structMap element for all files. This is a
        logical structMap that includes empty directories.
        """
        structmap = etree.Element(
            utils.lxmlns("mets") + "structMap",
            TYPE="logical",
            ID="structMap_2",
            LABEL="Normative Directory Structure",
        )
        for item in self._root_elements:
            child = item.serialize_structmap(recurse=True, normative=True)
            if child is not None:
                structmap.append(child)

        return structmap

    def _filesec(self, files=None):
        """
        Returns fileSec Element containing all files grouped by use.
        """
        if files is None:
            files = self.all_files()

        filesec = etree.Element(utils.lxmlns("mets") + "fileSec")
        filegrps = {}
        for file_ in files:
            if file_.type.lower() not in ("item", AIP_ENTRY_TYPE):
                continue
            # Get fileGrp, or create if not exist
            filegrp = filegrps.get(file_.use)
            if filegrp is None:
                filegrp = etree.SubElement(
                    filesec, utils.lxmlns("mets") + "fileGrp", USE=file_.use
                )
                filegrps[file_.use] = filegrp

            file_el = file_.serialize_filesec()
            if file_el is not None:
                filegrp.append(file_el)

        return filesec

    def serialize(self, fully_qualified=True):
        """
        Returns this document serialized to an xml Element.

        :return: Element for this document
        """
        now = datetime.utcnow().replace(microsecond=0).isoformat("T")
        files = self.all_files()
        mdsecs = self._collect_mdsec_elements(files)
        root = self._document_root(fully_qualified=fully_qualified)
        root.append(self._mets_header(now=now))
        for section in mdsecs:
            root.append(section.serialize(now=now))
        root.append(self._filesec(files))
        root.append(self._structmap())
        root.append(self._normative_structmap())
        return root

    def tostring(self, fully_qualified=True, pretty_print=True):
        """
        Serialize and return a string of this METS document.

        To write to file, see :meth:`write`

        :return: String of this document
        """
        root = self.serialize(fully_qualified=fully_qualified)
        return etree.tostring(root, pretty_print=pretty_print, xml_declaration=True)

    def write(self, filepath, fully_qualified=True, pretty_print=False):
        """Serialize and write this METS document to `filepath`.

        :param str filepath: Path to write the METS document to
        """
        root = self.serialize(fully_qualified=fully_qualified)
        tree = root.getroottree()
        tree.write(filepath, xml_declaration=True, pretty_print=pretty_print)

    # PARSE HELPERS

    def _parse_tree_structmap(self, tree, parent_elem, normative_parent_elem=None):
        """Recursively parse all the children of parent_elem, including amdSecs
        and dmdSecs.
        :param lxml._ElementTree tree: encodes the entire METS file.
        :param lxml._Element parent_elem: the element whose children we are
            parsing.
        :param lxml._Element normative_parent_elem: the normative
            counterpart of ``parent_elem`` taken from the logical structMap
            labelled "Normative Directory Structure".
        """
        siblings = []
        el_to_normative = self._get_el_to_normative(parent_elem, normative_parent_elem)
        for elem, normative_elem in el_to_normative.items():
            if elem.tag != utils.lxmlns("mets") + "div":
                continue  # Only handle divs, not fptrs
            entry_type = elem.get("TYPE")
            label = elem.get("LABEL")
            fptr_elems = elem.findall("mets:fptr", namespaces=utils.NAMESPACES)
            # Directories are walked recursively. Additionally, they may
            # contain direct fptrs.
            if entry_type.lower() == "directory":
                children = self._parse_tree_structmap(
                    tree, elem, normative_parent_elem=normative_elem
                )
                fs_entry = fsentry.FSEntry.dir(label, children)
                self._add_dmdsecs_to_fs_entry(elem, fs_entry, tree)
                siblings.append(fs_entry)
                for fptr_elem in fptr_elems:
                    fptr = self._analyze_fptr(fptr_elem, tree, entry_type)
                    fs_entry = fsentry.FSEntry.from_fptr(
                        label=None, type_=u"Item", fptr=fptr
                    )
                    self._add_amdsecs_to_fs_entry(fptr.amdids, fs_entry, tree)
                    siblings.append(fs_entry)
                continue
            # Other types, e.g.: items, aips...
            if not len(fptr_elems):
                continue
            fptr = self._analyze_fptr(fptr_elems[0], tree, entry_type)
            fs_entry = fsentry.FSEntry.from_fptr(label, entry_type, fptr)
            self._add_dmdsecs_to_fs_entry(elem, fs_entry, tree)
            self._add_amdsecs_to_fs_entry(fptr.amdids, fs_entry, tree)
            siblings.append(fs_entry)
        return siblings

    @staticmethod
    def _get_el_to_normative(parent_elem, normative_parent_elem):
        """Return ordered dict ``el_to_normative``, which maps children of
        ``parent_elem`` to their normative counterparts in the children of
        ``normative_parent_elem`` or to ``None`` if there is no normative
        parent. If there is a normative div element with no non-normative
        counterpart, that element is treated as a key with value ``None``.
        This allows us to create ``FSEntry`` instances for empty directory div
        elements, which are only documented in a normative logical structmap.
        """
        el_to_normative = OrderedDict()
        if normative_parent_elem is None:
            for el in parent_elem:
                el_to_normative[el] = None
        else:
            for norm_el in normative_parent_elem:
                matches = [
                    el
                    for el in parent_elem
                    if el.get("TYPE") == norm_el.get("TYPE")
                    and el.get("LABEL") == norm_el.get("LABEL")
                ]
                if matches:
                    el_to_normative[matches[0]] = norm_el
                else:
                    el_to_normative[norm_el] = None
        return el_to_normative

    @staticmethod
    def _analyze_fptr(fptr_elem, tree, entry_type):
        file_uuid = derived_from = use = path = amdids = checksum = checksumtype = None
        file_id = fptr_elem.get("FILEID")
        file_elem = tree.find(
            'mets:fileSec//mets:file[@ID="' + file_id + '"]',
            namespaces=utils.NAMESPACES,
        )
        if file_elem is None:
            raise exceptions.ParseError(
                "%s exists in structMap but not fileSec" % file_id
            )
        use = file_elem.getparent().get("USE")
        path = file_elem.find("mets:FLocat", namespaces=utils.NAMESPACES).get(
            utils.lxmlns("xlink") + "href"
        )
        try:
            path = utils.urldecode(path)
        except ValueError:
            raise exceptions.ParseError(
                'Value "{}" (of attribute xlink:href) is not a valid'
                " URL.".format(path)
            )
        amdids = file_elem.get("ADMID")
        checksum = file_elem.get("CHECKSUM")
        checksumtype = file_elem.get("CHECKSUMTYPE")
        file_id_prefix = utils.FILE_ID_PREFIX
        # If the file is an AIP, then its prefix is not "file-" but the
        # name of the AIP. Therefore we need to get the extension-less
        # basename of the AIP's path and remove its UUID suffix to ge
        # the prefix to remove from the FILEID attribute value.
        if entry_type.lower() == "archival information package":
            file_id_prefix = os.path.splitext(os.path.basename(path))[0][:-36]
        # If the file is part of a directory (with no intermediate item), then
        # its prefix *may not* be "file-" but the name of the file. This
        # pattern is found in old Archivematica METS files, e.g. see
        # ``fixtures/mets_dir_with_many_ptrs.xml``.
        elif entry_type.lower() == "directory" and file_id[:5] != "file-":
            file_id_prefix = os.path.basename(path) + "-"
        file_uuid = file_id.replace(file_id_prefix, "", 1)
        group_uuid = file_elem.get("GROUPID", "").replace(utils.GROUP_ID_PREFIX, "", 1)
        if group_uuid != file_uuid:
            derived_from = group_uuid  # Use group_uuid as placeholder
        return FPtr(file_uuid, derived_from, use, path, amdids, checksum, checksumtype)

    @staticmethod
    def _add_dmdsecs_to_fs_entry(elem, fs_entry, tree):
        dmdids = elem.get("DMDID")
        if dmdids:
            dmdids = dmdids.split()
            for dmdid in dmdids:
                dmdsec_elem = tree.find(
                    'mets:dmdSec[@ID="' + dmdid + '"]', namespaces=utils.NAMESPACES
                )
                dmdsec = metadata.SubSection.parse(dmdsec_elem)
                fs_entry.dmdsecs.append(dmdsec)
            # Create older/newer relationships
            fs_entry.dmdsecs.sort(key=lambda x: x.created)
            for prev_dmdsec, dmdsec in zip(fs_entry.dmdsecs, fs_entry.dmdsecs[1:]):
                if dmdsec.status == "updated":
                    prev_dmdsec.replace_with(dmdsec)

    @staticmethod
    def _add_amdsecs_to_fs_entry(amdids, fs_entry, tree):
        if amdids:
            amdids = amdids.split()
            for amdid in amdids:
                amdsec_elem = tree.find(
                    'mets:amdSec[@ID="' + amdid + '"]', namespaces=utils.NAMESPACES
                )
                amdsec = metadata.AMDSec.parse(amdsec_elem)
                fs_entry.amdsecs.append(amdsec)

    def _parse_tree(self, tree=None):
        if tree is None:
            tree = self.tree
        # self._validate()
        # Check CREATEDATE < now
        try:
            createdate = self.tree.find(
                "mets:metsHdr", namespaces=utils.NAMESPACES
            ).get("CREATEDATE")
        except AttributeError:
            createdate = None
        now = datetime.utcnow().isoformat("T")
        if createdate and createdate > now:
            raise exceptions.ParseError("CREATEDATE more recent than now (%s)" % now)
        self.createdate = createdate

        # Read root attributes
        root = tree
        if isinstance(tree, etree._ElementTree):
            root = tree.getroot()
        self.objid = root.get("OBJID", None)

        # Parse structMap
        structMap = tree.find(
            'mets:structMap[@TYPE="physical"]', namespaces=utils.NAMESPACES
        )
        if structMap is None:
            raise exceptions.ParseError("No physical structMap found.")
        normative_struct_map = tree.find(
            'mets:structMap[@TYPE="logical"]'
            '[@LABEL="Normative Directory Structure"]',
            namespaces=utils.NAMESPACES,
        )
        self._root_elements = self._parse_tree_structmap(
            tree, structMap, normative_parent_elem=normative_struct_map
        )

        # Associated derived files
        for entry in self.all_files():
            entry.derived_from = self.get_file(
                file_uuid=entry.derived_from, type="Item"
            )

    def _validate(self):
        raise NotImplementedError()

    @classmethod
    def fromfile(cls, path):
        """
        Creates a METS by parsing a file.

        :param str path: Path to a METS document.
        """
        parser = etree.XMLParser(remove_blank_text=True)

        return cls.fromtree(etree.parse(path, parser=parser))

    @classmethod
    def fromstring(cls, string):
        """
        Create a METS by parsing a string.

        :param str string: String containing a METS document.
        """
        parser = etree.XMLParser(remove_blank_text=True)
        root = etree.fromstring(string, parser)
        tree = root.getroottree()

        return cls.fromtree(tree)

    @classmethod
    def fromtree(cls, tree):
        """
        Create a METS from an ElementTree or Element.

        :param ElementTree tree: ElementTree to build a METS document from.
        """
        mets = cls()
        mets.tree = tree
        mets._parse_tree(tree)

        return mets


if __name__ == "__main__":
    mw = METSDocument()
    mw.fromfile(sys.argv[1])
    mw.write(sys.argv[2], fully_qualified=True, pretty_print=True)



# -*- coding: utf-8 -*-
from six.moves.urllib.parse import quote_plus, unquote_plus, urlparse, urlunparse


####################################
# LXML HELPER VALUES AND FUNCTIONS #
####################################

NAMESPACES = {
    "xsi": "http://www.w3.org/2001/XMLSchema-instance",
    "mets": "http://www.loc.gov/METS/",
    "premis": "info:lc/xmlns/premis-v2",
    "dcterms": "http://purl.org/dc/terms/",
    "fits": "http://hul.harvard.edu/ois/xml/ns/fits/fits_output",
    "xlink": "http://www.w3.org/1999/xlink",
    "dc": "http://purl.org/dc/elements/1.1/",
}

SCHEMA_LOCATIONS = (
    "http://www.loc.gov/METS/ "
    + "http://www.loc.gov/standards/mets/version111/mets.xsd"
)


def lxmlns(arg):
    """ Return XPath-usable namespace. """
    return "{" + NAMESPACES[arg] + "}"


####################
# METSRW CONSTANTS #
####################

FILE_ID_PREFIX = "file-"
GROUP_ID_PREFIX = "Group-"


#################################
# HELPERS FOR MANIPULATING URLS #
#################################

URL_ENCODABLE_PARTS = ("path", "params", "query", "fragment")


def _urlendecode(url, func):
    """Encode or decode ``url`` by applying ``func`` to all of its
    URL-encodable parts.
    """
    parsed = urlparse(url)
    for attr in URL_ENCODABLE_PARTS:
        parsed = parsed._replace(**{attr: func(getattr(parsed, attr))})
    return urlunparse(parsed)


def urlencode(url):
    """Replace unsafe ASCII characters using percent encoding as per RFC3986:
    https://tools.ietf.org/html/rfc3986#section-2.1.
    """
    return _urlendecode(url, lambda val: quote_plus(val, safe="/"))


def urldecode(url):
    """Decode percent encoding introduced per RFC3986
    https://tools.ietf.org/html/rfc3986#section-2.1.
    """
    return _urlendecode(url, unquote_plus)



# -*- coding: utf-8 -*-
"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

import codecs
from os import path
import re

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))


def read(*parts):
    with codecs.open(path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# Get the long description from the relevant file
with codecs.open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="metsrw",
    version=find_version("metsrw", "__init__.py"),
    description="Library for dealing with METS files.",
    long_description=long_description,
    url="https://github.com/artefactual-labs/mets-reader-writer/",
    author="Artefactual",
    author_email="info@artefactual.com",
    license="AGPL",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
    ],
    keywords="mets",
    packages=find_packages(exclude=["docs", "fixtures", "requirements", "tests*"]),
    install_requires=["future", "lxml", "six"],
    include_package_data=True,
    python_requires=">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*",
)



# -*- coding: utf-8 -*-
"""METS reader and writer."""

from __future__ import absolute_import
import logging

from .exceptions import MetsError, ParseError
from .fsentry import FSEntry
from .metadata import AMDSec, SubSection, MDRef, MDWrap
from .mets import METSDocument
from .utils import (
    NAMESPACES,
    SCHEMA_LOCATIONS,
    lxmlns,
    FILE_ID_PREFIX,
    GROUP_ID_PREFIX,
    urlencode,
    urldecode,
)
from .validate import (
    METS_XSD_PATH,
    AM_SCT_PATH,
    AM_PNTR_SCT_PATH,
    get_schematron,
    validate,
    get_xmlschema,
    xsd_validate,
    schematron_validate,
    sct_report_string,
    xsd_error_log_string,
    report_string,
)
from .di import (
    FeatureBroker,
    set_feature_broker_to_default_state,
    feature_broker,
    Dependency,
    has_class_methods,
    has_methods,
    is_class,
)
from . import plugins

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())
__version__ = "0.3.1"

__all__ = [
    "AMDSec",
    "AM_PNTR_SCT_PATH",
    "AM_SCT_PATH",
    "Dependency",
    "FILE_ID_PREFIX",
    "FSEntry",
    "FeatureBroker",
    "GROUP_ID_PREFIX",
    "MDRef",
    "MDWrap",
    "METSDocument",
    "METS_XSD_PATH",
    "MetsError",
    "NAMESPACES",
    "ParseError",
    "SCHEMA_LOCATIONS",
    "SubSection",
    "__version__",
    "feature_broker",
    "get_schematron",
    "get_xmlschema",
    "has_class_methods",
    "has_methods",
    "is_class",
    "lxmlns",
    "plugins",
    "report_string",
    "schematron_validate",
    "sct_report_string",
    "set_feature_broker_to_default_state",
    "urldecode",
    "urlencode",
    "validate",
    "xsd_error_log_string",
    "xsd_validate",
]



# -*- coding: utf-8 -*-
"""
Exceptions for metsrw.

All exceptions generated by this library will descend from MetsError.
"""


class MetsError(Exception):
    """ Base Exception for this module. """


class ParseError(MetsError):
    """ Error parsing a METS file. """


class SerializeError(MetsError):
    """ Error serializing a METS file. """
